/*
 * pajs_mixer.c
 *
 *  Created on: 10.07.2015
 *      Author: klausj
 */
#include "ips_audio_pulse_PulseMixerProvider.h"


JNIEXPORT jobject JNICALL Java_ips_audio_pulse_PulseMixerProvider_init(JNIEnv *env, jobject jobj,jstring applicationName,jstring iconName) {
	jobject nativePointer = NULL;
	jclass exceptionCls = (*env)->FindClass(env,
					"ips/audio/pulse/PulseException");
	if (exceptionCls == NULL) {
		fprintf(stderr,"Could not find exception class ips/audio/pulse/PulseException\n");
		fflush(stderr);
		return NULL;
	}
	JavaVM *javaVM;
	jint res = (*env)->GetJavaVM(env, &javaVM);
	if (res < 0) {
		(*env)->ThrowNew(env,exceptionCls,"Could not get Java VM");
	}else{
		PulseMixerProvider * pmp=(PulseMixerProvider*)malloc(sizeof(PulseMixerProvider));

		pmp->exceptionClassRef = (*env)->NewWeakGlobalRef(env, exceptionCls);
		pmp->javaVM = javaVM;
		pmp->javaObjRef = (*env)->NewWeakGlobalRef(env, jobj);

		nativePointer = (*env)->NewDirectByteBuffer(env, pmp,sizeof(PulseMixerProvider));
		jclass cls = (*env)->GetObjectClass(env, pmp->javaObjRef);
		if (cls == NULL) {
			(*env)->ThrowNew(env,exceptionCls,"Could not get object class");
		}
		pmp->jClassRef=(*env)->NewGlobalRef(env,cls);

		pa_threaded_mainloop *mainLoop=pa_threaded_mainloop_new();

		pa_mainloop_api* mainLoopApi=pa_threaded_mainloop_get_api(mainLoop);

		//pa_context* c=pa_context_new(mainLoopApi,"AJS");
		pa_proplist* p=pa_proplist_new();

		jboolean isCopy;
		const char *appName=NULL;
		const char *icName=NULL;
		if(applicationName!=NULL){
			appName = (*env)->GetStringUTFChars(env,applicationName, &isCopy);
			pa_proplist_set 	(p,
					PA_PROP_APPLICATION_NAME,
					appName,
					strlen(appName)+1
			);
		}else{
			appName="AJS";
		}
		if(iconName!=NULL){
			icName = (*env)->GetStringUTFChars(env,iconName, &isCopy);
			pa_proplist_set 	(p,
					PA_PROP_APPLICATION_ICON_NAME,
					icName,
					strlen(icName)+1
			);
		}
		// setting icon works as expected
		//		const char* appname="Speechrecorder";
		//		const char* icnname="speechrecorder";
		//		pa_proplist_set 	(p,
		//				PA_PROP_APPLICATION_NAME,
		//				appname,
		//				strlen(appname)+1
		//			);
		//		pa_proplist_set 	(p,
		//						PA_PROP_APPLICATION_ICON_NAME,
		//						icnname,
		//						strlen(icnname)+1
		//					);
		pa_context* c=pa_context_new_with_proplist(mainLoopApi,appName,p);
		pa_context_set_state_callback(c,context_state_callback,pmp);

		pmp->paCtx=c;
		pmp->paMainloop=mainLoop;
		if(applicationName!=NULL){
			(*env)->ReleaseStringUTFChars(env,applicationName,appName);
		}
		if(iconName!=NULL){
			(*env)->ReleaseStringUTFChars(env,iconName,icName);
		}
	}
	return nativePointer;
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_serverInfo(JNIEnv *env, jobject obj,jobject nativePointer) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,nativePointer);

	pa_operation* op=pa_context_get_server_info(pmp->paCtx,server_info_cb,pmp);

	if(op==NULL){
		(*env)->ThrowNew(env,pmp->exceptionClassRef,"PulseAudio: Server info query failed\n");
		return;
	}
}




JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_connect(JNIEnv *env, jobject obj,jobject nativePointer) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,nativePointer);

	int res=pa_context_connect(pmp->paCtx,NULL,0,NULL);

	if(res!=0){
		(*env)->ThrowNew(env,pmp->exceptionClassRef,"PulseAudio: Context connect failed\n");
		return;
	}
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_mainLoop(JNIEnv *env, jobject obj,jobject nativePointer) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,nativePointer);

	//pa_context_connect(pmp->paCtx,NULL,0,NULL);
	int res=pa_threaded_mainloop_start(pmp->paMainloop);
	if(res!=0){
		(*env)->ThrowNew(env,pmp->exceptionClassRef,"PulseAudio: Threaded mainloop start failed\n");
		return;
	}
}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_disconnect(JNIEnv *env, jobject obj,jobject nativePointer) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,nativePointer);
	pa_context_disconnect(pmp->paCtx);
}


JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_mainLoopQuit(JNIEnv *env, jobject obj,jobject nativePointer) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,nativePointer);
	pa_threaded_mainloop_stop(pmp->paMainloop);


}


JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_enumerateCards(JNIEnv *env, jobject obj,jobject nativePointer) {
	PulseMixerProvider* pmp = (*env)->GetDirectBufferAddress(env,nativePointer);

	pa_context_get_source_info_list(pmp->paCtx,source_info_list_cb,pmp);

	pa_context_get_sink_info_list(pmp->paCtx,sink_info_list_cb,pmp);


}

JNIEXPORT void JNICALL Java_ips_audio_pulse_PulseMixerProvider_release(JNIEnv *env, jobject jobj, jobject nativePointer) {

	PulseMixerProvider *pmp = (*env)->GetDirectBufferAddress(env,
			nativePointer);
	pa_context_unref(pmp->paCtx);
	(*env)->DeleteWeakGlobalRef(env, pmp->javaObjRef);
	(*env)->DeleteGlobalRef(env, pmp->jClassRef);
	free(pmp);
	//jnm_decodingplayer_release(jnmDecPlayer);
}

void context_state_callback(pa_context *context, void *userdata) {

	pa_context_state_t state = pa_context_get_state(context);
#ifdef DEBUG
	printf("PA: context state callback %d\n", state);
	fflush(stdout);
#endif
	if (state == PA_CONTEXT_READY || state == PA_CONTEXT_FAILED) {

		PulseMixerProvider *pmp = (PulseMixerProvider*) userdata;
		JNIEnv *env;
		JavaVM *jvm = pmp->javaVM;
		if (jvm != NULL) {
			jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
			if (res != JNI_OK) {
				fprintf(stderr, "Could not attach PA callback thread to JVM!");
				return;
			} else {
				//jclass cls = (*env)->GetObjectClass(env, pmp->javaObjRef);
				//if (cls == NULL) {
				//	fprintf(stderr,"Could not get object class\n");
				//	return;
				//}
				jmethodID mId = NULL;
				char* mName;
				if (state == PA_CONTEXT_READY) {
#ifdef DEBUG
					printf("PA: Context ready\n");
#endif
					mName = "contextReady";
				} else if (state == PA_CONTEXT_FAILED) {
#ifdef DEBUG
					printf("PA: Context failed\n");
#endif
					mName = "contextFailed";
				} else if (state == PA_CONTEXT_TERMINATED) {
					mName = "contextTerminated";
				}
				mId = (*env)->GetMethodID(env, pmp->jClassRef, mName, "()V");
				if (mId == NULL) {
					fprintf(stderr, "Method not found: %s",mName);
					return;
				}

				(*env)->CallVoidMethod(env, pmp->javaObjRef, mId);
				(*jvm)->DetachCurrentThread(jvm);
			}
		}
	}else{
#ifdef DEBUG
		printf("PA: context state callback %d\n", state);
		fflush(stdout);
		return;
#endif
}
}

void server_info_cb(pa_context *c, const pa_server_info *i, void *userdata){
	PulseMixerProvider *pmp=(PulseMixerProvider*)userdata;
		JNIEnv *env;
		JavaVM *jvm = pmp->javaVM;
		if (jvm != NULL) {
			jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
			if (res != JNI_OK) {
				fprintf(stderr,"Could not attach PA callback thread to JVM!");
				return;
			}else{
				jmethodID mId=NULL;
				mId=(*env)->GetMethodID(env,pmp->jClassRef,"serverInfo","(Ljava/lang/String;Ljava/lang/String;)V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					fflush(stderr);
					return;
				}
				jstring defSourceNm=(*env)->NewStringUTF(env,i->default_source_name);
				jstring defSinkNm=(*env)->NewStringUTF(env,i->default_sink_name);
				(*env)->CallVoidMethod(env,pmp->javaObjRef,mId,defSourceNm,defSinkNm);
				(*jvm)->DetachCurrentThread(jvm);
			}
		}
}

void card_info_list_cb(pa_context *c, const pa_card_info *i, int eol, void *userdata){
#ifdef DEBUG
	if(!eol){
		printf("PA: card info: driver: %s, name: %s, index: %d\n",i->driver,i->name,i->index);
		// TODO build JS Mixer.Info
		pa_card_port_info** pis=i->ports;
		while(*pis!=NULL){
			pa_card_port_info* pi=*pis;
			printf("PA: port info: name: %s\n",pi->name);
			pis++;
		}
	}
#endif
}

void source_info_list_cb(pa_context *c, const pa_source_info *i, int eol, void *userdata){

	PulseMixerProvider *pmp=(PulseMixerProvider*)userdata;
	JNIEnv *env;
	JavaVM *jvm = pmp->javaVM;
	if (jvm != NULL) {
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res != JNI_OK) {
			fprintf(stderr,"Could not attach PA callback thread to JVM!");
			return;
		}else{
			jmethodID mId=NULL;
			if(eol){
				mId=(*env)->GetMethodID(env,pmp->jClassRef,"sourceInfoListEnd","()V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					return;
				}
				(*env)->CallVoidMethod(env,pmp->javaObjRef,mId);
			}else{
#ifdef DEBUG
				printf("PA: source driver: %s, name:  %s, descr: %s, card idx: %d,monitor of sink %d\n",i->driver,i->name,i->description,i->card,i->monitor_of_sink);
#endif
				mId=(*env)->GetMethodID(env,pmp->jClassRef,"sourceInfo","(ILjava/lang/String;Ljava/lang/String;)V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					return;
				}
				jstring sinkName=(*env)->NewStringUTF(env,i->name);
				jstring sinkDescr=(*env)->NewStringUTF(env,i->description);

				pa_format_info** fmtInfos=i->formats;
				uint8_t j;
				jboolean pcmSupported=JNI_FALSE;
				for(j=0;j<i->n_formats;j++){
					pa_format_info* fmtInfo=fmtInfos[j];
					if(!pcmSupported && fmtInfo->encoding==PA_ENCODING_PCM){
						pcmSupported=JNI_TRUE;
					}

#ifdef DEBUG
					printf("PA: fmt info: encoding %d\n",fmtInfo->encoding);
					printf("PA: fmt info is Any PCM format %d\n",fmtInfo->encoding==PA_ENCODING_PCM);
					printf("PA: fmt info: prop empty ?  %d\n",pa_proplist_isempty(fmtInfo->plist));
					printf("PA: fmt info: props:  %s\n",pa_proplist_to_string(fmtInfo->plist));
					void* is=NULL;
					const char* key=NULL;
					do{
						key=pa_proplist_iterate 	(fmtInfo->plist,&is);
						if(key!=NULL){
							printf("PA: fmt info: prop:  %s\n",key);
						}
					}while(key!=NULL);
#endif
				}

				// Do not show monitors of sinks (other applications such as the Ubuntu audio control panel don't do as well)
				// ( if monitor_of_sink is a (non negative) index it is a monitor
				if(pcmSupported && i->monitor_of_sink==PA_INVALID_INDEX){
					(*env)->CallVoidMethod(env,pmp->javaObjRef,mId,i->index,sinkName,sinkDescr);
				}
			}
			(*jvm)->DetachCurrentThread(jvm);
		}
	}
#ifdef DEBUG
	fflush(stdout);
#endif
}

void sink_info_list_cb(pa_context *c, const pa_sink_info *i, int eol, void *userdata){
	PulseMixerProvider *pmp=(PulseMixerProvider*)userdata;
	JNIEnv *env;
	JavaVM *jvm = pmp->javaVM;
	if (jvm != NULL) {
		jint res = (*jvm)->AttachCurrentThread(jvm, (void **) &env, NULL);
		if (res != JNI_OK) {
					fprintf(stderr,"Could not attach PA callback thread to JVM!");
					return;
				}else{


			jmethodID mId=NULL;
			if(eol){
				mId=(*env)->GetMethodID(env,pmp->jClassRef,"sinkInfoListEnd","()V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					return;
				}
				(*env)->CallVoidMethod(env,pmp->javaObjRef,mId);
			}else{
#ifdef DEBUG
				printf("PA: sink driver: %s, name:  %s, descr: %s, card idx: %d\n",i->driver,i->name,i->description,i->card);
#endif
				mId=(*env)->GetMethodID(env,pmp->jClassRef,"sinkInfo","(ILjava/lang/String;Ljava/lang/String;)V");
				if(mId==NULL){
					fprintf(stderr,"Method not found!");
					return;
				}
				jstring sinkName=(*env)->NewStringUTF(env,i->name);
				jstring sinkDescr=(*env)->NewStringUTF(env,i->description);

				pa_format_info** fmtInfos=i->formats;
				uint8_t j;
				jboolean pcmSupported=JNI_FALSE;
				for(j=0;j<i->n_formats;j++){
					pa_format_info* fmtInfo=fmtInfos[j];
					if(!pcmSupported && fmtInfo->encoding==PA_ENCODING_PCM){
						pcmSupported=JNI_TRUE;
					}
#ifdef DEBUG
					printf("PA: fmt info: encoding %d\n",fmtInfo->encoding);
					printf("PA: fmt info is Any PCM format %d\n",fmtInfo->encoding==PA_ENCODING_PCM);
					printf("PA: fmt info: prop empty ?  %d\n",pa_proplist_isempty(fmtInfo->plist));
					printf("PA: fmt info: props:  %s\n",pa_proplist_to_string(fmtInfo->plist));
					void* is=NULL;
					const char* key=NULL;
					do{
						key=pa_proplist_iterate 	(fmtInfo->plist,&is);
						if(key!=NULL){
							printf("PA: fmt info: prop:  %s\n",key);
						}
					}while(key!=NULL);
#endif
				}
				(*env)->CallVoidMethod(env,pmp->javaObjRef,mId,i->index,sinkName,sinkDescr);
			}
			(*jvm)->DetachCurrentThread(jvm);
		}
	}
#ifdef DEBUG
	fflush(stdout);
#endif
}












